/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JContainer3.java 871 2006-07-13 14:23:56Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.container;

import static org.objectweb.easybeans.container.mdb.MDBMessageEndPointFactory.DEFAULT_ACTIVATION_SPEC_NAME;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.resource.ResourceException;
import javax.resource.spi.ActivationSpec;
import javax.resource.spi.ResourceAdapter;

import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.EZBContainerCallbackInfo;
import org.objectweb.easybeans.api.EZBContainerConfig;
import org.objectweb.easybeans.api.EZBContainerException;
import org.objectweb.easybeans.api.EZBContainerLifeCycleCallback;
import org.objectweb.easybeans.api.EZBServer;
import org.objectweb.easybeans.api.Factory;
import org.objectweb.easybeans.api.FactoryException;
import org.objectweb.easybeans.container.info.MessageDrivenInfo;
import org.objectweb.easybeans.container.info.SessionBeanInfo;
import org.objectweb.easybeans.container.mdb.MDBMessageEndPointFactory;
import org.objectweb.easybeans.container.mdb.MDBResourceAdapterHelper;
import org.objectweb.easybeans.container.session.SessionFactory;
import org.objectweb.easybeans.container.session.stateful.StatefulSessionFactory;
import org.objectweb.easybeans.container.session.stateless.StatelessSessionFactory;
import org.objectweb.easybeans.deployment.Deployment;
import org.objectweb.easybeans.deployment.annotations.exceptions.AnalyzerException;
import org.objectweb.easybeans.deployment.annotations.exceptions.ResolverException;
import org.objectweb.easybeans.deployment.annotations.impl.JLocal;
import org.objectweb.easybeans.deployment.annotations.impl.JRemote;
import org.objectweb.easybeans.deployment.annotations.metadata.ClassAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.EjbJarAnnotationMetadata;
import org.objectweb.easybeans.deployment.helper.JavaContextHelper;
import org.objectweb.easybeans.deployment.helper.JavaContextHelperException;
import org.objectweb.easybeans.deployment.resolver.JNDIResolver;
import org.objectweb.easybeans.deployment.xml.EJB3DeploymentDescException;
import org.objectweb.easybeans.enhancer.Enhancer;
import org.objectweb.easybeans.enhancer.EnhancerException;
import org.objectweb.easybeans.jmx.MBeansException;
import org.objectweb.easybeans.jmx.MBeansHelper;
import org.objectweb.easybeans.loader.EasyBeansClassLoader;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.persistence.PersistenceUnitManager;
import org.objectweb.easybeans.persistence.api.EZBPersistenceUnitManager;
import org.objectweb.easybeans.persistence.xml.PersistenceXmlFileAnalyzer;
import org.objectweb.easybeans.persistence.xml.PersistenceXmlFileAnalyzerException;
import org.objectweb.easybeans.rpc.LocalCallRef;
import org.objectweb.easybeans.rpc.RemoteCallRef;
import org.objectweb.easybeans.server.Embedded;

/**
 * Defines an EJB3 container.
 * @author Florent Benoit
 */
public class JContainer3 implements EZBContainer {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(JContainer3.class);

    /**
     * Link to the EasyBeans server of this container.
     */
    private EZBServer embedded = null;

    /**
     * Id of this container.
     */
    private String id = null;

    /**
     * Archive used by this container.
     */
    private File archive = null;

    /**
     * Classloader use to manage this archive.
     */
    private ClassLoader classLoader = null;

    /**
     * Deployment for the managed archive.
     */
    private Deployment deployment = null;

    /**
     * Container available.
     */
    private boolean available = false;

    /**
     * Map of managed ejb3 factories.
     */
    private Map<String, Factory> factories = null;

    /**
     * Persistence manager object which manages all persistence-unit associated
     * to this container.
     */
    private EZBPersistenceUnitManager persistenceUnitManager = null;

    /**
     * Callbacks.
     */
    private List<EZBContainerLifeCycleCallback> callbacksLifeCycle = null;

    /**
     * JContainer Configuration.
     */
    private EZBContainerConfig configuration = null;

    /**
     * Build a new container on the given archive.
     * @param config The JContainer configuration storing the archive (jar file
     *        or exploded).
     */
    public JContainer3(final EZBContainerConfig config) {
        this.configuration = config;
        this.archive = this.configuration.getArchive();
        this.embedded = this.configuration.getEZBServer();
        this.id = String.valueOf(System.identityHashCode(this));
        deployment = new Deployment(archive);
        this.factories = new HashMap<String, Factory>();
        callbacksLifeCycle = this.configuration.getCallbacks();
    }

    /**
     * Gets the id of this container.
     * @return string id.
     */
    public String getId() {
        return id;
    }

    /**
     * Start this container.
     * @throws EZBContainerException if starting fails.
     */
    public void start() throws EZBContainerException {

        // Analyze files
        long tStart = System.currentTimeMillis();
        try {
            deployment.analyze();
        } catch (AnalyzerException e) {
            throw new EZBContainerException("Cannot analyze archive '" + archive.getName() + "'.", e);
        } catch (ResolverException e) {
            throw new EZBContainerException("Cannot resolve some annotations in the archive '" + archive.getName()
                    + "'.", e);
        } catch (EJB3DeploymentDescException e) {
            throw new EZBContainerException("Cannot parse deployment descriptor in the archive '" + archive.getName()
                    + "'.", e);
        }
        if (logger.isInfoEnabled()) {
            logger.info("Analyze elapsed during : " + (System.currentTimeMillis() - tStart) + " ms");
        }

        // Run enhancer on classes found
        URL url;
        try {
            url = archive.toURL();
        } catch (MalformedURLException e) {
            throw new EZBContainerException("Cannot get URL from archive '" + archive.getName() + "'.", e);
        }

        // Build JNDI resolver
        JNDIResolver jndiResolver = new JNDIResolver();
        jndiResolver.addDeployment(deployment);

        // add resolver in the map
        Map<String, Object> enhancerMap = new HashMap<String, Object>();
        enhancerMap.put(JNDIResolver.NAME, jndiResolver);

        ClassLoader old = Thread.currentThread().getContextClassLoader();
        // Define classloader if it was not yet defined (ear case -->
        // classloader already set)
        if (classLoader == null) {
            classLoader = new EasyBeansClassLoader(new URL[] {url}, old);
        }
        try {
            Thread.currentThread().setContextClassLoader(classLoader);
            Enhancer enhancer = new Enhancer(classLoader, deployment.getAnnotationDeploymentAnalyzer()
                    .getEjbJarAnnotationMetadata(), enhancerMap);

            long tStartEnhancing = System.currentTimeMillis();
            try {
                enhancer.enhance();
            } catch (EnhancerException e) {
                throw new EZBContainerException("Cannot run enhancer on archive '" + archive.getName() + "'.", e);
            }
            if (logger.isInfoEnabled()) {
                logger.info("Enhancement elapsed during : " + (System.currentTimeMillis() - tStartEnhancing) + " ms");
            }

            // Check if there is META-INF/persistence.xml file
            PersistenceUnitManager analyzedPersistenceUnitManager = null;
            try {
                analyzedPersistenceUnitManager = PersistenceXmlFileAnalyzer.analyzePersistenceXmlFile(archive,
                        getClassLoader());
            } catch (PersistenceXmlFileAnalyzerException e) {
                throw new EZBContainerException("Cannot analyze the persistence.xml file in the archive", e);
            }

            // No previous manager
            if (this.persistenceUnitManager == null) {
                this.persistenceUnitManager = analyzedPersistenceUnitManager;
            } else {
                // merge old and new.
                if (analyzedPersistenceUnitManager != null) {
                    analyzedPersistenceUnitManager.merge(persistenceUnitManager);
                    // update persistence manager with the merged one.
                    this.persistenceUnitManager = analyzedPersistenceUnitManager;
                }
            }

            // Create Beans Factories
            createBeanFactories();

            // cleanup
            deployment = new Deployment(archive);
            enhancer = null;
        } finally {
            Thread.currentThread().setContextClassLoader(old);
        }
        if (logger.isInfoEnabled()) {
            logger.info("Container started in : " + (System.currentTimeMillis() - tStart) + " ms");
        }
        available = true;

        // Send notification to callbacks
        if (callbacksLifeCycle.size() > 0) {
            EZBContainerCallbackInfo info = getContainer3CallbackInfo();
            for (EZBContainerLifeCycleCallback callback : callbacksLifeCycle) {
                try {
                    callback.start(info);
                } catch (Throwable t) {
                    // Protection from malicious code
                    logger.error("{0}.start() failed", callback.getClass().getName(), t);
                }
            }
        }

    }

    /**
     * Create the factories of the beans (session and MDB).
     * @throws EZBContainerException if binding fails.
     */
    private void createBeanFactories() throws EZBContainerException {
        // bind session beans
        EjbJarAnnotationMetadata ejbMetadata = deployment.getAnnotationDeploymentAnalyzer()
                .getEjbJarAnnotationMetadata();
        if (ejbMetadata != null) {
            for (ClassAnnotationMetadata bean : ejbMetadata.getClassAnnotationMetadataCollection()) {
                Factory factory = null;
                if (bean.isSession()) {
                    factory = createSessionBeanFactory(bean);
                } else if (bean.isMdb()) {
                    factory = createMessageDrivenBeanFactory(bean);
                }

                // Post-Configure the created factories.
                if (factory != null) {
                    // Build java: Context
                    Context javaContext;
                    try {
                        javaContext = JavaContextHelper.build(bean, factory);
                    } catch (JavaContextHelperException e) {
                        throw new EZBContainerException("Cannot build environment", e);
                    }

                    // Set java: context
                    factory.setJavaContext(javaContext);

                    // Add Management
                    // TODO: always done ! add a switch !
                    try {
                        MBeansHelper.getInstance().registerMBean(factory);
                    } catch (MBeansException me) {
                        throw new EZBContainerException("Cannot register the factory MBean", me);
                    }


                    // Init factory
                    try {
                        factory.init();
                    } catch (FactoryException e) {
                        throw new EZBContainerException("Cannot initialize the factory.", e);
                    }
                }
            }
        }
    }

    /**
     * Creates the given message driven bean factory.
     * @param messageDrivenBean the message driven bean class metadata.
     * @throws EZBContainerException if the message driven bean cannot be
     *         created.
     * @return the build factory.
     */
    private Factory createMessageDrivenBeanFactory(final ClassAnnotationMetadata messageDrivenBean)
            throws EZBContainerException {
        String className = messageDrivenBean.getClassName().replace("/", ".");

        // get Activation spec object
        ActivationSpec activationSpec = null;
        try {
            activationSpec = (ActivationSpec) new InitialContext().lookup(DEFAULT_ACTIVATION_SPEC_NAME);
        } catch (NamingException e1) {
            throw new EZBContainerException("Cannot get the activation spec with the name '"
                    + DEFAULT_ACTIVATION_SPEC_NAME + "'.");
        }

        // get ResourceAdapter object
        ResourceAdapter resourceAdapter = null;
        try {
            resourceAdapter = MDBResourceAdapterHelper.getResourceAdapter(DEFAULT_ACTIVATION_SPEC_NAME,
                    (Embedded) getConfiguration().getEZBServer());
        } catch (ResourceException e) {
            throw new EZBContainerException("Cannot get the resource adapter for this MDB factory", e);
        }

        // Associate Resource Adapter with ActivationSpec object (if not set)
        if (activationSpec.getResourceAdapter() == null) {
            try {
                activationSpec.setResourceAdapter(resourceAdapter);
            } catch (ResourceException e) {
                throw new EZBContainerException("Cannot associate resource adapter with activation spec object", e);
            }
        }

        // Create factory
        MDBMessageEndPointFactory mdbMessageEndPointFactory = null;
        try {
            mdbMessageEndPointFactory = new MDBMessageEndPointFactory(className, this, activationSpec, resourceAdapter);
        } catch (FactoryException e) {
            throw new EZBContainerException("Cannot build the MDB MessageEndPoint factory", e);
        }

        // build runtime information
        MessageDrivenInfo messageDrivenInfo = new MessageDrivenInfo();
        messageDrivenInfo.setApplicationExceptions(messageDrivenBean.getEjbJarAnnotationMetadata().getApplicationExceptions());
        messageDrivenInfo.setTransactionManagementType(messageDrivenBean.getTransactionManagementType());
        messageDrivenInfo.setMessageListenerInterface(messageDrivenBean.getJMessageDriven()
                .getMessageListenerInterface());
        messageDrivenInfo.setActivationConfigProperties(messageDrivenBean.getJMessageDriven()
                .getActivationConfigProperties());
        mdbMessageEndPointFactory.setMessageDrivenInfo(messageDrivenInfo);



        return mdbMessageEndPointFactory;

    }

    /**
     * Creates the given session bean and bind it.
     * @param sessionBean the session bean class metadata.
     * @throws EZBContainerException if the session bean cannot be created
     * @return the build factory.
     */
    private Factory createSessionBeanFactory(final ClassAnnotationMetadata sessionBean) throws EZBContainerException {
        String className = sessionBean.getClassName().replace("/", ".");

        SessionFactory sessionFactory = null;

        if (sessionBean.isStateless()) {
            try {
                sessionFactory = new StatelessSessionFactory(className, this);
            } catch (FactoryException fe) {
                throw new EZBContainerException("Cannot build the stateless factory", fe);
            }
        } else {
            try {
                sessionFactory = new StatefulSessionFactory(className, this);
            } catch (FactoryException fe) {
                throw new EZBContainerException("Cannot build the stateless factory", fe);
            }
        }

        // build runtime information
        SessionBeanInfo sessionBeanInfo = new SessionBeanInfo();
        sessionBeanInfo.setTransactionManagementType(sessionBean.getTransactionManagementType());
        sessionBeanInfo.setApplicationExceptions(sessionBean.getEjbJarAnnotationMetadata().getApplicationExceptions());
        sessionFactory.setSessionBeanInfo(sessionBeanInfo);

        // Add the factory to the managed factories
        factories.put(className, sessionFactory);

        // get interface of bean
        JLocal localItfs = sessionBean.getLocalInterfaces();
        JRemote remoteItfs = sessionBean.getRemoteInterfaces();

        if (localItfs != null) {
            for (String itf : localItfs.getInterfaces()) {
                bindLocalItf(itf, embedded.getID(), getId(), className, sessionBean, "Local");
            }
        }
        if (remoteItfs != null) {
            for (String itf : remoteItfs.getInterfaces()) {
                bindRemoteItf(itf, getId(), className, sessionBean, "Remote");
            }
        }

        return sessionFactory;
    }

    /**
     * Stop this container.
     */
    public void stop() {
        available = false;

        // stop each factories
        for (Factory f : factories.values()) {
            f.stop();
        }

        // reset CL
        classLoader = null;

        // TODO : no unbind for now
        // need to configure it ?

        // Send notification to callbacks
        if (callbacksLifeCycle.size() > 0) {
            EZBContainerCallbackInfo info = getContainer3CallbackInfo();
            for (EZBContainerLifeCycleCallback callback : callbacksLifeCycle) {
                try {
                    callback.stop(info);
                } catch (Throwable t) {
                    // Protection from malicious code
                    logger.error("{0}.stop() failed", callback.getClass().getName(), t);
                }
            }
        }
    }

    /**
     * Gets information on the container that can be given to container
     * callbacks.
     * @return information on the managed container.
     */
    private EZBContainerCallbackInfo getContainer3CallbackInfo() {
        EZBContainerCallbackInfo info = new EZBContainerCallbackInfo();
        info.setArchive(archive);
        info.setFactories(factories);
        return info;
    }

    /**
     * Bind into registry interfaces for a bean making them available from
     * clients.
     * @param itf the name of the interface that object will have.
     * @param embeddedId the ID of the embedded server.
     * @param containerID the ID of the container.
     * @param factoryName the name of the factory.
     * @param bean the bean class associated to given interface.
     * @param mode Local, Remote, etc
     * @throws EZBContainerException if interface cannot be loaded or if the
     *         bind fails
     */
    private void bindLocalItf(final String itf, final Integer embeddedId, final String containerID, final String factoryName,
            final ClassAnnotationMetadata bean, final String mode) throws EZBContainerException {
        String itfClsName = itf.replace("/", ".");
        Class clz = null;
        try {
            clz = classLoader.loadClass(itfClsName);
        } catch (ClassNotFoundException e) {
            throw new EZBContainerException(
                    "Cannot find the class '" + itf + "' in Classloader '" + classLoader + "'.", e);
        }

        // Build a reference
        LocalCallRef localCall = new LocalCallRef(itfClsName, embeddedId, containerID, factoryName, bean.isStateful());

        try {
            String jndiName = null;
            String mappedName = bean.getJCommonBean().getMappedName();
            if (mappedName != null && !"".equals(mappedName)) {
                jndiName = mappedName;
            } else {
                jndiName = jndiNameEncode(bean.getClassName(), itfClsName, mode);
            }
            logger.debug("Binding bean {0} with interface {1} into registry with jndi name {2}", bean.getClassName()
                    .replace("/", "."), itf.replace("/", "."), jndiName);
            new InitialContext().rebind(jndiName, localCall);
        } catch (NamingException e) {
            throw new EZBContainerException("Cannot bind the object '" + localCall + "' into JNDI with the name '"
                    + clz.getName() + "'.", e);
        }

    }

    /**
     * Bind into registry interfaces for a bean making them available from
     * clients.
     * @param itf the name of the interface that object will have.
     * @param containerID the ID of the container.
     * @param factoryName the name of the factory.
     * @param bean the bean class associated to given interface.
     * @param mode Local, Remote, etc
     * @throws EZBContainerException if interface cannot be loaded or if the
     *         bind fails
     */
    private void bindRemoteItf(final String itf, final String containerID, final String factoryName,
            final ClassAnnotationMetadata bean, final String mode) throws EZBContainerException {
        String itfClsName = itf.replace("/", ".");
        Class clz = null;
        try {
            clz = classLoader.loadClass(itfClsName);
        } catch (ClassNotFoundException e) {
            throw new EZBContainerException(
                    "Cannot find the class '" + itf + "' in Classloader '" + classLoader + "'.", e);
        }

        // Build a reference
        RemoteCallRef remoteCall = new RemoteCallRef(itfClsName, containerID, factoryName, bean.isStateful());

        try {
            String jndiName = null;
            String mappedName = bean.getJCommonBean().getMappedName();
            if (mappedName != null && !"".equals(mappedName)) {
                jndiName = mappedName;
            } else {
                jndiName = jndiNameEncode(bean.getClassName(), itfClsName, mode);
            }
            logger.debug("Binding bean {0} with interface {1} into registry with jndi name {2}", bean.getClassName()
                    .replace("/", "."), itf.replace("/", "."), jndiName);
            new InitialContext().rebind(jndiName, remoteCall);
        } catch (NamingException e) {
            throw new EZBContainerException("Cannot bind the object '" + remoteCall + "' into JNDI with the name '"
                    + clz.getName() + "'.", e);
        }

    }

    /**
     * Gets a factory with its given name.
     * @param factoryName the factory name.
     * @return the factory found or null.
     */
    public Factory getFactory(final String factoryName) {
        return factories.get(factoryName);
    }

    /**
     * @return Returns a Collection of managed Factories.
     */
    public Collection<Factory> getFactories() {
        return factories.values();
    }

    /**
     * Gets the name of this container.
     * @return the name.
     */
    public String getName() {
        return archive.getName();
    }

    /**
     * Gets the classloader.
     * @return classloader of the container
     */
    public ClassLoader getClassLoader() {
        return classLoader;
    }

    /**
     * Gets the archive used by this container. It can be a .jar file or a
     * directory.
     * @return the archive.
     */
    public File getArchive() {
        return archive;
    }

    /**
     * Check if the container is available or not.
     * @return true if the container is available.
     */
    public boolean isAvailable() {
        return available;
    }

    /**
     * Gets the persistence manager object which manages all persistence-unit
     * associated to this container.
     * @return persistence unit manager object
     */
    public EZBPersistenceUnitManager getPersistenceUnitManager() {
        return persistenceUnitManager;
    }

    /**
     * Encode a given name by prefixing it with given mode.
     * @param beanClassName name of the bean class.
     * @param itfName name of the interface.
     * @param mode local/remote, etc
     * @return encoded jndi Name
     */
    public static String jndiNameEncode(final String beanClassName, final String itfName, final String mode) {
        return beanClassName.replace("/", ".") + "_" + itfName.replace("/", ".") + "@" + mode;
    }

    /**
     * Sets the classloader.
     * @param classLoader to be used by the container
     */
    public void setClassLoader(final ClassLoader classLoader) {
        if (this.classLoader != null) {
            throw new IllegalArgumentException("Cannot replace an existing classloader");
        }
        this.classLoader = classLoader;
    }

    /**
     * Sets the persistence manager object which manages all persistence-unit
     * associated to this container.
     * @param persistenceUnitManager persistence unit manager object to set.
     */
    public void setPersistenceUnitManager(final EZBPersistenceUnitManager persistenceUnitManager) {
        if (this.persistenceUnitManager != null) {
            throw new IllegalArgumentException("Cannot replace an existing persistenceUnitManager");
        }
        this.persistenceUnitManager = persistenceUnitManager;
    }

    /**
     * @return Returns the Container Configuration.
     */
    public EZBContainerConfig getConfiguration() {
        return this.configuration;
    }

}
