/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MDBMessageEndPointFactory.java 675 2006-06-18 18:16:49Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.container.mdb;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import javax.ejb.ActivationConfigProperty;
import javax.resource.ResourceException;
import javax.resource.spi.ActivationSpec;
import javax.resource.spi.InvalidPropertyException;
import javax.resource.spi.ResourceAdapter;
import javax.resource.spi.UnavailableException;
import javax.resource.spi.endpoint.MessageEndpoint;
import javax.resource.spi.endpoint.MessageEndpointFactory;
import javax.transaction.xa.XAResource;

import org.objectweb.easybeans.api.EZBContainer;
import org.objectweb.easybeans.api.FactoryException;
import org.objectweb.easybeans.api.bean.EasyBeansMDB;
import org.objectweb.easybeans.pool.PoolException;

/**
 * Defines a class that will manage the message end point factory for the MDB.
 * The super class will manage the pool of message end point.
 * @author Florent Benoit
 */
public class MDBMessageEndPointFactory extends MDBFactory implements MessageEndpointFactory {

    /**
     * Default name of the activation spec (JORAM).
     */
    public static final String DEFAULT_ACTIVATION_SPEC_NAME = "joramActivationSpec";

    /**
     * ActivationSpec object linked to this factory (used to activate or
     * deactive an endpoint factory (us).
     */
    private ActivationSpec activationSpec = null;

    /**
     * Resource adapter that provides the activation spec implementation.
     */
    private ResourceAdapter resourceAdapter = null;

    /**
     * Default constructor (delegate to super class).
     * @param className name of this factory (name of class that is managed)
     * @param container the root component of this factory.
     * @param activationSpec the activation Spec object used for
     *        activating/deactivating.
     * @param resourceAdapter the resource adapter used to activate/deactivate
     *        ourself.
     * @throws FactoryException if super constructor fails
     */
    public MDBMessageEndPointFactory(final String className, final EZBContainer container,
            final ActivationSpec activationSpec, final ResourceAdapter resourceAdapter) throws FactoryException {
        super(className, container);
        this.activationSpec = activationSpec;
        this.resourceAdapter = resourceAdapter;
    }

    /**
     * Init the factory.
     * @throws FactoryException if the initialization fails.
     */
    @Override
    public void init() throws FactoryException {
        initActivationSpec();

        validateActivationSpec();

        activate();
    }


    /**
     * Call setters method on the activation spec object.
     * @throws FactoryException if activation spec object is not configured.
     */
    private void initActivationSpec() throws FactoryException {
        // Call setter method for each property found in the activation
        // properties
        List<ActivationConfigProperty> properties = getMessageDrivenInfo().getActivationConfigProperties();
        if (properties != null) {
            for (ActivationConfigProperty property : properties) {
                // get property name
                String key = property.propertyName();
                // and value
                String value = property.propertyValue();

                // define setter method name
                String methodName = "set" + key.substring(0, 1).toUpperCase() + key.substring(1);

                // get Method (reflection)
                Method m = null;
                try {
                    m = activationSpec.getClass().getMethod(methodName, new Class[] {String.class});
                } catch (SecurityException e) {
                    throw new FactoryException("Cannot get a method named '" + methodName
                            + "' on activation spec object '" + activationSpec + "'.", e);
                } catch (NoSuchMethodException e) {
                    throw new FactoryException("Cannot get a method named '" + methodName
                            + "' on activation spec object '" + activationSpec + "'.", e);
                }

                // invoke method
                try {
                    m.invoke(activationSpec, value);
                } catch (IllegalArgumentException e) {
                    throw new FactoryException("Cannot invoke method named '" + methodName + "' with value '" + value
                            + "' on activation spec object '" + activationSpec + "'.", e);
                } catch (IllegalAccessException e) {
                    throw new FactoryException("Cannot invoke method named '" + methodName + "' with value '" + value
                            + "' on activation spec object '" + activationSpec + "'.", e);
                } catch (InvocationTargetException e) {
                    throw new FactoryException("Cannot invoke method named '" + methodName + "' with value '" + value
                            + "' on activation spec object '" + activationSpec + "'.", e);
                }

            }
        }

    }

    /**
     * Validate the configuration used p, tje activation spec object.
     * @throws FactoryException if the validation of the activation spec
     *         implementation object fails.
     */
    private void validateActivationSpec() throws FactoryException {
        try {
            activationSpec.validate();
        } catch (InvalidPropertyException e) {
            throw new FactoryException("Cannot validate the validation spec object", e);
        }
    }

    /**
     * Activate this endpoint factory on resource adapter with the activation
     * spec object.
     * @throws FactoryException if the activation fails.
     */
    private void activate() throws FactoryException {
        try {
            resourceAdapter.endpointActivation(this, activationSpec);
        } catch (ResourceException e) {
            throw new FactoryException(
                    "Cannot activate the activationspec object and us (MessageEndPointFactory) on the resource adapter",
                    e);
        }
    }

    /**
     * This is used to create a message endpoint. The message endpoint is
     * expected to implement the correct message listener type.
     * @param xaResource an optional XAResource instance used to get transaction
     *        notifications when the message delivery is transacted.
     * @return a message endpoint instance.
     * @throws UnavailableException indicates a transient failure in creating a
     *         message endpoint. Subsequent attempts to create a message
     *         endpoint might succeed.
     */
    public MessageEndpoint createEndpoint(final XAResource xaResource) throws UnavailableException {
        // Instance that implements the MessageEndpoint interface
        MDBMessageEndPoint messageEndpoint = null;

        // get an instance of MDB
        EasyBeansMDB easyBeansMDB = null;
        try {
            easyBeansMDB = getPool().get();
        } catch (PoolException e) {
            throw new UnavailableException("Cannot get instance in the pool", e);
        }

        // Build a wrapper around this mdb instance
        // TODO: For now, this is only for JMS, needs to be changed.
        messageEndpoint = new MDBMessageListenerEndPoint(this, easyBeansMDB);

        // Set XAResource of the message endpoint.
        messageEndpoint.setXaResource(xaResource);

        return messageEndpoint;
    }

    /**
     * Release an endpoint created by this factory.
     * @param mdbMessageEndPoint the endpoint to release.
     */
    protected void releaseEndPoint(final MDBMessageEndPoint mdbMessageEndPoint) {
        // Release the wrapped message driven bean
        try {
            getPool().release(mdbMessageEndPoint.getEasyBeansMDB());
        } catch (PoolException e) {
            throw new IllegalStateException("Cannot release the given message end point", e);
        }
    }

    /**
     * This is used to find out whether message deliveries to a target method on
     * a message listener interface that is implemented by a message endpoint
     * will be transacted or not. The message endpoint may indicate its
     * transacted delivery preferences (at a per method level) through its
     * deployment descriptor. The message delivery preferences must not change
     * during the lifetime of a message endpoint.
     * @param method description of a target method. This information about the
     *        intended target method allows an application server to find out
     *        whether the target method call will be transacted or not.
     * @return boolean whether the specified method is transacted
     * @throws NoSuchMethodException exception to throw
     */
    public boolean isDeliveryTransacted(final Method method) throws NoSuchMethodException {
        // TODO : Not yet implemented
        return false;
    }

    /**
     * Stops the factory.
     */
    @Override
    public void stop() {
        // stop the pool.
        super.stop();

        // deactivate this factory
        resourceAdapter.endpointDeactivation(this, activationSpec);

    }
}
