/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MDBResourceAdapterHelper.java 675 2006-06-18 18:16:49Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.container.mdb;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import javax.resource.ResourceException;
import javax.resource.spi.ResourceAdapter;

import org.objectweb.easybeans.jms.joram.JoramService;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.server.Embedded;

/**
 * This class allow to get the ResourceAdapter object for a given destination
 * (activation-spec object).
 * @author Florent Benoit
 */
public final class MDBResourceAdapterHelper {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(MDBResourceAdapterHelper.class);

    /**
     * Utility class, no public constructor.
     */
    private MDBResourceAdapterHelper() {

    }

    /**
     * Gets the resource adapter object for the given jndi name (activation
     * spec) and the given embedded object.
     * @param jndiName the nameof the activation spec bound in the registry
     * @param embedded the embedded server
     * @return an instance of the resource adapter that provides the MDB
     *         activation spec.
     * @throws ResourceException if an error occurs while trying to get the
     *         resource adapter.
     */
    public static ResourceAdapter getResourceAdapter(final String jndiName, final Embedded embedded)
            throws ResourceException {
        // try to see if JMS mini-resource adapter service was started.
        JoramService joramService = embedded.getJoramService();
        if (joramService != null) {
            // get ResourceAdapter from the service
            return joramService.getResourceAdapter();
        }

        // else, maybe we are in JOnAS ? try to get the class with reflection
        String jRoot = System.getProperty("install.root");
        if (jRoot != null) {
            logger.debug("JOnAS detected, trying with JOnAS Resource Adapter");
            return getJOnASResourceAdapter(jndiName);
        }

        throw new ResourceException("MDB is used but no JMS resource service was started."
                + "Also, no resource service found for the server.");

    }

    /**
     * Gets the ResourceAdapter by using JOnAS code (used when embedded into
     * JOnAS application server).
     * @param jndiName the name bound in the registry for activation spec
     *        object.
     * @return an instance of the resource adapter.
     * @throws ResourceException if there is a failure when retrieving the
     *         resource adapter object.
     */
    public static ResourceAdapter getJOnASResourceAdapter(final String jndiName) throws ResourceException {
        Class rarClass = null;
        try {
            rarClass = Thread.currentThread().getContextClassLoader().loadClass("org.objectweb.jonas.resource.Rar");
        } catch (ClassNotFoundException e) {
            throw new ResourceException("Cannot find the JOnAS resource adapter class", e);
        }

        // Get method for Rar.getRar(jndiName);
        Method getRarMethod = null;
        try {
            getRarMethod = rarClass.getMethod("getRar", new Class[] {String.class});
        } catch (SecurityException e) {
            throw new ResourceException("Cannot get the getRar method on the class '" + rarClass + "'.", e);
        } catch (NoSuchMethodException e) {
            throw new ResourceException("Cannot get the getRar method on the class '" + rarClass + "'.", e);
        }

        // invoke method (static method)
        Object rarObject = null;
        try {
            rarObject = getRarMethod.invoke(null, jndiName);
        } catch (IllegalArgumentException e) {
            throw new ResourceException("Cannot invoke method with jndiName '" + jndiName + "'.", e);
        } catch (IllegalAccessException e) {
            throw new ResourceException("Cannot invoke method with jndiName '" + jndiName + "'.", e);
        } catch (InvocationTargetException e) {
            throw new ResourceException("Cannot invoke method with jndiName '" + jndiName + "'.", e
                    .getTargetException());
        }

        // get the resource adapter on the given object
        Method getResourceAdapterMethod = null;
        try {
            getResourceAdapterMethod = rarObject.getClass().getMethod("getResourceAdapter", new Class[] {});
        } catch (SecurityException e) {
            throw new ResourceException("Cannot get the getResourceAdapter method on the class '"
                    + rarObject.getClass() + "'.", e);
        } catch (NoSuchMethodException e) {
            throw new ResourceException("Cannot get the getResourceAdapter method on the class '"
                    + rarObject.getClass() + "'.", e);
        }

        // invoke method
        Object resourceAdapterObj = null;
        try {
            resourceAdapterObj = getResourceAdapterMethod.invoke(rarObject, new Object[] {});
        } catch (IllegalArgumentException e) {
            throw new ResourceException("Cannot invoke method getResourceAdapter on the rar object.", e);
        } catch (IllegalAccessException e) {
            throw new ResourceException("Cannot invoke method getResourceAdapter on the rar object.", e);
        } catch (InvocationTargetException e) {
            throw new ResourceException("Cannot invoke method getResourceAdapter on the rar object.", e);
        }

        // cast object
        ResourceAdapter resourceAdapter = null;
        if (resourceAdapterObj instanceof ResourceAdapter) {
            resourceAdapter = (ResourceAdapter) resourceAdapterObj;
        } else {
            throw new ResourceException("Object found is not an instance of ResourceAdapter");
        }
        return resourceAdapter;

    }
}
