/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AnnotationDeploymentAnalyzer.java 340 2006-04-14 17:29:44Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployment.annotations.analyzer;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import org.objectweb.asm.ClassReader;

import org.objectweb.easybeans.deployment.annotations.exceptions.AnalyzerException;
import org.objectweb.easybeans.deployment.annotations.metadata.EjbJarAnnotationMetadata;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * This class finds the annotated class and fill metadata class.
 * @author Florent Benoit
 */
public class AnnotationDeploymentAnalyzer {

    /**
     * Archive which will be analyzed.
     */
    private File archive = null;

    /**
     * Metadata representing the parsed ejb-jar file.
     */
    private EjbJarAnnotationMetadata ejbJarAnnotationMetadata = null;

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(AnnotationDeploymentAnalyzer.class);

    /**
     * ASM visitor used to visit classes.
     */
    private ScanClassVisitor scanVisitor = null;

    /**
     * Constructor.<br>
     * Archive which will be used when analyzing.
     * @param archive the archive to analyze.
     */
    public AnnotationDeploymentAnalyzer(final File archive) {
        this.archive = archive;
        this.ejbJarAnnotationMetadata = new EjbJarAnnotationMetadata();
        scanVisitor = new ScanClassVisitor(ejbJarAnnotationMetadata);
    }

    /**
     * Analyzes the archive and fill metadata struct.
     * @throws AnalyzerException if analyze of archive fails
     */
    public void analyze() throws AnalyzerException {

        // two cases :
        // - archive is a jar file
        // - archive is an exploded directory

        if (archive.isFile()) {
            scanJarArchive();
        } else {
            // exploded mode
            // need recursion on all directories of the archive
            scanDirectory(archive);
        }
    }

    /**
     * Scan all classes from the archive which is a jar file.
     * @throws AnalyzerException if scan is aborted.
     */
    private void scanJarArchive() throws AnalyzerException {
        JarFile jarFile;
        try {
            jarFile = new JarFile(archive);
        } catch (IOException ioe) {
            throw new AnalyzerException("Cannot build jar file on archive '" + archive + "'.", ioe);
        }
        Enumeration<? extends ZipEntry> en = jarFile.entries();
        while (en.hasMoreElements()) {
            ZipEntry e = en.nextElement();
            String name = e.getName();
            if (name.toLowerCase().endsWith(".class")) {
                if (logger.isDebugEnabled()) {
                    logger.debug("Visiting class '" + name + " from jar-file '" + jarFile.getName());
                }
                try {
                    new ClassReader(jarFile.getInputStream(e)).accept(scanVisitor, false);
                } catch (Exception ioe) {
                    throw new AnalyzerException("Error while analyzing file entry '" + name + "' in jar file '"
                            + archive + "'", ioe);
                }

            }
        }
        try {
            jarFile.close();
        } catch (IOException ioe) {
            logger.warn("Error while trying to close the file '" + jarFile + "'", ioe);
        }

    }

    /**
     * Scan (recursively) all classes from the given directory. It starts with the archive root directory first.
     * @param directory the directory in which proceed all classes.
     * @throws AnalyzerException if analyzer fails to scan classes.
     */
    private void scanDirectory(final File directory) throws AnalyzerException {
        File[] files = directory.listFiles();
        for (File f : files) {
            // scan .class file
            if (f.isFile() && f.getName().toLowerCase().endsWith(".class")) {
                InputStream is = null;
                try {
                    is = new FileInputStream(f);
                } catch (FileNotFoundException e) {
                    throw new AnalyzerException("Cannot read input stream from '" + f.getAbsolutePath() + "'.", e);
                }
                try {
                    new ClassReader(is).accept(scanVisitor, false);
                } catch (Exception e) {
                    throw new AnalyzerException("Cannot launch class visitor on the file '" + f.getAbsolutePath() + "'.", e);
                }
                try {
                    is.close();
                } catch (IOException e) {
                    throw new AnalyzerException("Cannot close input stream on the file '" + f.getAbsolutePath() + "'.", e);
                }
            } else {
                // loop on childs
                if (f.isDirectory()) {
                    scanDirectory(f);
                }
            }
        }
    }

    /**
     * @return struct (metadata) filled by this analyzer
     */
    public EjbJarAnnotationMetadata getEjbJarAnnotationMetadata() {
        return ejbJarAnnotationMetadata;
    }

}
