/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MetadataMerge.java 537 2006-05-29 17:05:14Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployment.xml.helper;

import static org.objectweb.easybeans.deployment.xml.struct.LifeCycleCallback.POST_CONSTRUCT;
import static org.objectweb.easybeans.deployment.xml.struct.LifeCycleCallback.PRE_DESTROY;

import java.util.List;

import org.objectweb.easybeans.deployment.annotations.ClassType;
import org.objectweb.easybeans.deployment.annotations.impl.JCommonBean;
import org.objectweb.easybeans.deployment.annotations.impl.JInterceptors;
import org.objectweb.easybeans.deployment.annotations.impl.JLocal;
import org.objectweb.easybeans.deployment.annotations.impl.JRemote;
import org.objectweb.easybeans.deployment.annotations.metadata.ClassAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.EjbJarAnnotationMetadata;
import org.objectweb.easybeans.deployment.annotations.metadata.MethodAnnotationMetadata;
import org.objectweb.easybeans.deployment.xml.struct.AbsBean;
import org.objectweb.easybeans.deployment.xml.struct.EJB3;
import org.objectweb.easybeans.deployment.xml.struct.EnterpriseBeans;
import org.objectweb.easybeans.deployment.xml.struct.LifeCycleCallback;
import org.objectweb.easybeans.deployment.xml.struct.Session;

/**
 * This class merge the information of the struct representing EJB DD into the
 * metadata.
 * @author Florent Benoit
 */
public final class MetadataMerge {

    /**
     * Metadata for an ejb-jar.
     */
    private EjbJarAnnotationMetadata ejbMetadata = null;

    /**
     * Helper class, no public constructor.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    private MetadataMerge(final EjbJarAnnotationMetadata ejbMetadata) {
        this.ejbMetadata = ejbMetadata;
    }

    /**
     * Takes struct of metadata and adds/set information on class/methods
     * metadata.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    public static void merge(final EjbJarAnnotationMetadata ejbMetadata) {
        new MetadataMerge(ejbMetadata).resolve();
    }

    /**
     * Do all merging operations.
     */
    private void resolve() {

        EJB3 ejb3 = ejbMetadata.getEjb3();
        if (ejb3 != null) {
            // Get enterprise beans
            EnterpriseBeans enterpriseBeans = ejb3.getEnterpriseBeans();
            if (enterpriseBeans != null) {

                // get session beans
                List<Session> sessionList = enterpriseBeans.getSessionList();
                for (Session session : sessionList) {
                    // get ejb-name
                    String ejbName = session.getEjbName();
                    String ejbClass = session.getEjbClass();
                    ClassAnnotationMetadata classAnnotationMetadata = findClassForEjb(ejbName, ejbClass);

                    // class is found, then apply all customization.
                    applySessionBean(session, classAnnotationMetadata);
                }
            }
        }
    }

    /**
     * Apply rules for a session bean.
     * @param sessionbean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applySessionBean(final Session sessionbean, final ClassAnnotationMetadata classAnnotationMetadata) {
        applyCommonBean(sessionbean, classAnnotationMetadata);

        // set session-type
        String sessionType = sessionbean.getSessionType();
        if (sessionType == null && !classAnnotationMetadata.isSession()) {
            throw new IllegalStateException("Missing session-type for bean '" + sessionbean
                    + "' and no annotation in '" + classAnnotationMetadata + "'.");
        }
        // set the type of bean
        if ("Stateless".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.STATELESS);
        } else if ("Stateful".equals(sessionType)) {
            classAnnotationMetadata.setClassType(ClassType.STATEFUL);
        }

        // local interface
        if (sessionbean.getBusinessLocalList().size() > 0) {
            JLocal jLocal = classAnnotationMetadata.getLocalInterfaces();
            if (jLocal == null) {
                jLocal = new JLocal();
                classAnnotationMetadata.setLocalInterfaces(jLocal);
            }
            for (String itf : sessionbean.getBusinessLocalList()) {
                String encodedItf = encode(itf);
                if (!jLocal.contains(encodedItf)) {
                    jLocal.addInterface(encodedItf);
                }
            }
        }

        // remote interface
        if (sessionbean.getBusinessRemoteList().size() > 0) {
            JRemote jRemote = classAnnotationMetadata.getRemoteInterfaces();
            if (jRemote == null) {
                jRemote = new JRemote();
                classAnnotationMetadata.setRemoteInterfaces(jRemote);
            }
            for (String itf : sessionbean.getBusinessRemoteList()) {
                String encodedItf = encode(itf);
                if (!jRemote.contains(encodedItf)) {
                    jRemote.addInterface(encodedItf);
                }
            }
        }
    }

    /**
     * Apply rules for a common bean.
     * @param bean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyCommonBean(final AbsBean bean, final ClassAnnotationMetadata classAnnotationMetadata) {
        // no Common bean, create one.
        if (classAnnotationMetadata.getJCommonBean() == null) {
            classAnnotationMetadata.setJCommonBean(new JCommonBean());
        }
        // Sets the ejb-name
        JCommonBean jcommonBean = classAnnotationMetadata.getJCommonBean();
        jcommonBean.setName(bean.getEjbName());

        // TODO: ordering ?
        // post-construct / pre-destroy
        applyLifeCycle(bean, bean.getPostConstructList(), classAnnotationMetadata, POST_CONSTRUCT);
        applyLifeCycle(bean, bean.getPreDestroyList(), classAnnotationMetadata, PRE_DESTROY);

    }

    /**
     * Apply rules for a common bean (lifecycle callbacks).
     * @param bean the struct (XML DD)
     * @param lifecycleList the list of lifecycle callbacks.
     * @param classAnnotationMetadata (Annotation metadata).
     * @param type the type of lifecycle.
     */
    private void applyLifeCycle(final AbsBean bean, final List<LifeCycleCallback> lifecycleList,
            final ClassAnnotationMetadata classAnnotationMetadata, final String type) {
        for (LifeCycleCallback lifecycleCallback : lifecycleList) {

            String encodedClassName = null;

            // if no class, takes the bean's class
            if (lifecycleCallback.getClassName() == null) {
                encodedClassName = classAnnotationMetadata.getClassName();
            } else {
                encodedClassName = encode(lifecycleCallback.getClassName());
            }
            JInterceptors interceptors = classAnnotationMetadata.getAnnotationInterceptors();
            if (interceptors == null) {
                interceptors = new JInterceptors();
                classAnnotationMetadata.setAnnotationsInterceptors(interceptors);
            }
            if (!interceptors.contains(encodedClassName)) {
                interceptors.addClass(encodedClassName);
            }
            // add LifeCycle on the metadata class
            ClassAnnotationMetadata lifeCycleMetadata = ejbMetadata.getClassAnnotationMetadata(encodedClassName);
            if (lifeCycleMetadata == null) {
                throw new IllegalStateException("Cannot find metadata for class '" + encodedClassName
                        + "' which is in a lifecycle.");
            }

            // get method and call setter to true
            for (MethodAnnotationMetadata methodMetadata : lifeCycleMetadata.getMethodAnnotationMetadataCollection()) {
                if (methodMetadata.getMethodName().equals(lifecycleCallback.getMethod())) {
                    if (POST_CONSTRUCT.equals(type)) {
                        methodMetadata.setPostConstruct(true);
                        lifeCycleMetadata.addPostConstructMethodMetadata(methodMetadata);
                    } else if (PRE_DESTROY.equals(type)) {
                        methodMetadata.setPreDestroy(true);
                        lifeCycleMetadata.addPreDestroyMethodMetadata(methodMetadata);
                    } else {
                        throw new IllegalStateException("No case for type '" + type + "'..");
                    }
                }
            }
        }
    }

    /**
     * Find/Gets a classAnnotationMetadata for a given ejb name and/or a given
     * ejbclass name.
     * @param ejbName the name of the EJB.
     * @param ejbClass the class of the EJB.
     * @return a classAnnotationMetadata if found, else exception
     */
    private ClassAnnotationMetadata findClassForEjb(final String ejbName, final String ejbClass) {
        ClassAnnotationMetadata classAnnotationMetadata = null;

        // First, try with ejbClass if not null
        if (ejbClass != null) {
            classAnnotationMetadata = ejbMetadata.getClassAnnotationMetadata(encode(ejbClass));
            // found ? return it
            if (classAnnotationMetadata != null) {
                return classAnnotationMetadata;
            }
        }

        // else, needs to find a class with the specified name
        for (ClassAnnotationMetadata tmpMetaData : ejbMetadata.getClassAnnotationMetadataCollection()) {
            if (tmpMetaData.isBean()) {
                // check name
                JCommonBean bean = tmpMetaData.getJCommonBean();
                if (bean != null) {
                    // found !
                    if (ejbName != null && ejbName.equals(bean.getName())) {
                        return tmpMetaData;
                    }
                }
            }
        }
        throw new IllegalStateException("No class with ejb-name '" + ejbName + "' or ejb-class '" + ejbClass
                + "' was found in the EJB-JAR file '" + ejbMetadata);
    }

    /**
     * Encode a classname (replace . by /) as it's the internal form of the
     * metadata
     * @param className the name of the class to encode
     * @return encoded name
     */
    private static String encode(final String className) {
        return className.replace(".", "/");
    }
}
