/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EJB3DeploymentDescLoader.java 519 2006-05-28 19:49:27Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.deployment.xml.parsing;

import static org.objectweb.easybeans.deployment.xml.struct.LifeCycleCallback.POST_CONSTRUCT;
import static org.objectweb.easybeans.deployment.xml.struct.LifeCycleCallback.PRE_DESTROY;

import java.net.URL;
import java.util.List;

import org.objectweb.easybeans.deployment.xml.struct.AbsBean;
import org.objectweb.easybeans.deployment.xml.struct.EJB3;
import org.objectweb.easybeans.deployment.xml.struct.EnterpriseBeans;
import org.objectweb.easybeans.deployment.xml.struct.LifeCycleCallback;
import org.objectweb.easybeans.deployment.xml.struct.Session;
import org.objectweb.easybeans.deployment.xml.struct.common.AbsEnvironment;
import org.objectweb.easybeans.deployment.xml.struct.common.EnvEntry;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;
import org.objectweb.easybeans.util.xml.DocumentParser;
import org.objectweb.easybeans.util.xml.DocumentParserException;
import org.objectweb.easybeans.util.xml.XMLUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * This class analyzes the EJB3 deployment descriptor and fill a struct with
 * getters.
 * @author Florent Benoit
 */
public final class EJB3DeploymentDescLoader {

    /**
     * Persistence namespace.
     */
    private static final String JAVAEE_NS = "http://java.sun.com/xml/ns/javaee";

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(EJB3DeploymentDescLoader.class);

    /**
     * Validating with schema ?
     */
    private static boolean validating = true;

    /**
     * Utility class, no constructor.
     */
    private EJB3DeploymentDescLoader() {

    }

    /**
     * Load the ejb-jar.xml file.
     * @param url the URL of the XML file.
     * @throws ParsingException if parsing of XML file fails.
     * @return an application object.
     */
    public static EJB3 loadDeploymentDescriptor(final URL url) throws ParsingException {
        logger.debug("Analyzing url {0}", url);

        // Get document
        Document document = null;
        try {
            document = DocumentParser.getDocument(url, validating, new EJB3EntityResolver());
        } catch (DocumentParserException e) {
            throw new ParsingException("Cannot parse the url", e);
        }

        EJB3 ejb3 = new EJB3();

        // Root element = <ejb-jar>
        Element ejbJarRootElement = document.getDocumentElement();

        NodeList enterprisebeansList = ejbJarRootElement.getElementsByTagNameNS(JAVAEE_NS, EnterpriseBeans.NAME);
        if (enterprisebeansList.getLength() == 1) {
            Element enterpriseBeansElement = (Element) enterprisebeansList.item(0);

            // Create struct
            EnterpriseBeans enterpriseBeans = new EnterpriseBeans();
            ejb3.setEnterpriseBeans(enterpriseBeans);

            // Analyze beans
            analyzeBeans(enterpriseBeansElement, enterpriseBeans);

        }

        return ejb3;

    }

    /**
     * Analyze common stuff of beans.
     * @param enterpriseBeansElement the dom element.
     * @param enterpriseBeans the structure representing object.
     */
    private static void analyzeBeans(final Element enterpriseBeansElement, final EnterpriseBeans enterpriseBeans) {
        // Get session beans
        NodeList sessionBeansList = enterpriseBeansElement.getElementsByTagNameNS(JAVAEE_NS, Session.NAME);

        // Loop on session beans
        for (int i = 0; i < sessionBeansList.getLength(); i++) {
            Element sessionElement = (Element) sessionBeansList.item(i);

            // Build instance that is created.
            Session session = new Session();

            enterpriseBeans.addSession(session);

            // First, analyze common things
            analyzeSessionBean(sessionElement, session);
        }

    }

    /**
     * Analyze common stuff of a bean.
     * @param beanElement the dom element.
     * @param bean the structure representing object.
     */
    private static void analyzeBean(final Element beanElement, final AbsBean bean) {
        // Get ejb-name
        String ejbName = XMLUtils.getStringValueElement(JAVAEE_NS, beanElement, "ejb-name");
        bean.setEjbName(ejbName);

        // Get ejb-class
        String ejbClass = XMLUtils.getStringValueElement(JAVAEE_NS, beanElement, "ejb-class");
        bean.setEjbClass(ejbClass);

        // Analyze common things
        analyzeEnvironement(beanElement, bean);
    }

    /**
     * Analyze XML of a session bean.
     * @param beanElement the dom element.
     * @param bean the structure representing object.
     */
    private static void analyzeSessionBean(final Element beanElement, final Session bean) {
        analyzeBean(beanElement, bean);

        // Get business-remote
        List<String> businessRemoteList = XMLUtils.getStringListValueElement(JAVAEE_NS, beanElement, "business-remote");
        for (String businessRemote : businessRemoteList) {
            bean.addBusinessRemote(businessRemote);
        }

        // Get business-local
        List<String> businessLocalList = XMLUtils.getStringListValueElement(JAVAEE_NS, beanElement, "business-local");
        for (String businessLocal : businessLocalList) {
            bean.addBusinessLocal(businessLocal);
        }

        // Get session-type
        String sessionType = XMLUtils.getStringValueElement(JAVAEE_NS, beanElement, "session-type");
        bean.setSessionType(sessionType);

        // Get transaction-type
        String transactionType = XMLUtils.getStringValueElement(JAVAEE_NS, beanElement, "transaction-type");
        bean.setTransactionType(transactionType);

    }

    /**
     * Analyze the environment (could be a bean, servlet, etc).
     * @param envElement the dom element.
     * @param bean the structure representing object.
     */
    private static void analyzeEnvironement(final Element envElement, final AbsEnvironment bean) {
        analyzeLifeCycle(envElement, bean, POST_CONSTRUCT);
        analyzeLifeCycle(envElement, bean, PRE_DESTROY);
        analyzeEnvEnvtry(envElement, bean);
    }

    /**
     * Analyze the post-construct and pre-destroy element.
     * @param envElement the dom element.
     * @param bean the structure representing object.
     * @param lifecycleName the name of lifecycle (post-construct, etc)
     */
    private static void analyzeLifeCycle(final Element envElement, final AbsEnvironment bean, final String lifecycleName) {
        // Get the lifecycle element
        NodeList lifecycleList = envElement.getElementsByTagNameNS(JAVAEE_NS, lifecycleName);

        // Loop on the lifecycle
        for (int i = 0; i < lifecycleList.getLength(); i++) {
            Element lifeCycleElement = (Element) lifecycleList.item(i);

            // Build instance that is created.
            LifeCycleCallback lifecycleCallback = new LifeCycleCallback();
            if (POST_CONSTRUCT.equals(lifecycleName)) {
                bean.addPostConstruct(lifecycleCallback);
            } else if (PRE_DESTROY.equals(lifecycleName)) {
                bean.addPreDestroy(lifecycleCallback);
            } else {
                throw new IllegalArgumentException("Unknown lifecycle named '" + lifecycleName + "'.");
            }

            // lifecycle-callback-class
            String className = XMLUtils.getStringValueElement(JAVAEE_NS, lifeCycleElement, "lifecycle-callback-class");
            lifecycleCallback.setClassName(className);

            // lifecycle-callback-method
            String method = XMLUtils.getStringValueElement(JAVAEE_NS, lifeCycleElement, "lifecycle-callback-method");
            lifecycleCallback.setMethod(method);

        }

    }

    /**
     * Analyze the env-entry element.
     * @param envElement the dom element.
     * @param bean the structure representing object.
     */
    private static void analyzeEnvEnvtry(final Element envElement, final AbsEnvironment bean) {
        // Get env-entry
        NodeList envEntryList = envElement.getElementsByTagNameNS(JAVAEE_NS, EnvEntry.NAME);

        // Loop on env entry
        for (int i = 0; i < envEntryList.getLength(); i++) {
            Element envEntryElement = (Element) envEntryList.item(i);

            // Build instance that is created.
            EnvEntry envEntry = new EnvEntry();
            bean.addEnvEntry(envEntry);

            // Env-entry name
            String envEntryName = XMLUtils.getStringValueElement(JAVAEE_NS, envEntryElement, "env-entry-name");
            envEntry.setEnvEntryName(envEntryName);

            // Env-entry type
            String envEntryType = XMLUtils.getStringValueElement(JAVAEE_NS, envEntryElement, "env-entry-type");
            envEntry.setEnvEntryType(envEntryType);

            // Env-entry value
            String envEntryValue = XMLUtils.getStringValueElement(JAVAEE_NS, envEntryElement, "env-entry-value");
            envEntry.setEnvEntryValue(envEntryValue);
        }
    }

}
