/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: TomcatENCInterceptor.java 872 2006-07-13 17:18:15Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.objectweb.easybeans.naming.interceptors;

import java.lang.reflect.Method;

import javax.naming.Context;

import org.objectweb.easybeans.api.EasyBeansInterceptor;
import org.objectweb.easybeans.api.EasyBeansInvocationContext;
import org.objectweb.easybeans.log.JLog;
import org.objectweb.easybeans.log.JLogFactory;

/**
 * Interceptor used when EasyBeans is integrated in Tomcat. As the java:
 * namespace is managed by Tomcat, EasyBeans needs to call Tomcat objects in
 * order to set java: context.
 * @author Florent Benoit
 */
public class TomcatENCInterceptor implements EasyBeansInterceptor {

    /**
     * Logger.
     */
    private static JLog logger = JLogFactory.getLog(TomcatENCInterceptor.class);

    /**
     * Tomcat's class for naming.
     */
    protected static final String TOMCAT_NAMING_CLASS = "org.apache.naming.ContextBindings";

    /**
     * Method used to bind the name associated to a component context for the current thread on the Tomcat naming.
     */
    private static Method bindThreadMethod = null;

    /**
     * Method used to unbind the name associated to a context for the current thread on the Tomcat naming.
     */
    private static Method unbindThreadMethod = null;

    /**
     * Method used to bind the component context for the current thread on the Tomcat naming.
     */
    private static Method bindContextMethod = null;

    /**
     * Method used to unbind the component context for the current thread on the Tomcat naming.
     */
    private static Method unbindContextMethod = null;


    /**
     * Default constructor. Gets a reference on the tomcat's naming methods
     */
    public TomcatENCInterceptor() {
        // Get reference on naming manager and its methods.
        if (bindThreadMethod == null) {
            String errMsg = "Check that EasyBeans is embedded in Tomcat web server.";
            Class namingClass = null;
            try {
                namingClass = Thread.currentThread().getContextClassLoader().loadClass(TOMCAT_NAMING_CLASS);
            } catch (ClassNotFoundException e) {
                throw new IllegalStateException("Cannot load the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            }
            // get method (bindThread)
            try {
                bindThreadMethod = namingClass.getMethod("bindThread", new Class[] {Object.class});
            } catch (SecurityException e) {
                throw new IllegalStateException("Cannot get bindThread() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            } catch (NoSuchMethodException e) {
                throw new IllegalStateException("Cannot get bindThread() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            }

            // get method (unbindThread)
            try {
                unbindThreadMethod = namingClass.getMethod("unbindThread", new Class[] {Object.class});
            } catch (SecurityException e) {
                throw new IllegalStateException("Cannot get unbindThread() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            } catch (NoSuchMethodException e) {
                throw new IllegalStateException("Cannot get unbindThread() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            }

            // get method (bindContext)
            try {
                bindContextMethod = namingClass.getMethod("bindContext", new Class[] {Object.class, Context.class});
            } catch (SecurityException e) {
                throw new IllegalStateException("Cannot get bindContext() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            } catch (NoSuchMethodException e) {
                throw new IllegalStateException("Cannot get bindContext() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            }

            // get method (unbindContext)
            try {
                unbindContextMethod = namingClass.getMethod("unbindContext", new Class[] {Object.class});
            } catch (SecurityException e) {
                throw new IllegalStateException("Cannot get unbindContext() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            } catch (NoSuchMethodException e) {
                throw new IllegalStateException("Cannot get unbindContext() method on the Tomcat naming class '"
                        + TOMCAT_NAMING_CLASS + "'. " + errMsg, e);
            }

        }
    }

    /**
     * Sets JOnAS ENC context.
     * @param invocationContext context with useful attributes on the current
     *        invocation.
     * @return result of the next invocation (to chain interceptors).
     * @throws Exception needs for signature of interceptor.
     */
    public Object intercept(final EasyBeansInvocationContext invocationContext) throws Exception {
        bindContextMethod.invoke(null, "EZBContext", invocationContext.getFactory().getJavaContext());
        bindThreadMethod.invoke(null, "EZBContext");
        try {
            return invocationContext.proceed();
        } finally {
            logger.debug("Unset ENC environment");
            unbindThreadMethod.invoke(null, "EZBContext");
            unbindContextMethod.invoke(null, "EZBContext");
        }
    }

}
